<?php

namespace QUI\ERP\Payments\Crypto\ETH;

use QUI;
use QUI\ERP\Accounting\Payments\Api\AbstractPaymentProvider;
use QUI\ERP\Accounting\Payments\Transactions\Handler as TransactionsHandler;
use QUI\ERP\Payments\Crypto\BlockchainTransaction;
use QUI\ERP\Payments\Crypto\CryptoPaymentBlockchainProviderInterface;
use QUI\ERP\Payments\Crypto\CryptoPaymentType;
use QUI\ERP\Payments\Crypto\Exception\TransactionNotFoundException;
use QUI\ERP\Payments\Crypto\Utils as CryptoPaymentsUtils;

use function hexdec;
use function json_decode;
use function mb_strlen;
use function mb_substr;

/**
 * Class BlockchainProvider
 */
class BlockchainProvider implements CryptoPaymentBlockchainProviderInterface
{
    protected static string $etherscanBaseUrl = 'https://api.etherscan.io/api';
    protected static string $etherscanRopstenBaseUrl = 'https://api-ropsten.etherscan.io/api';
    protected static string $etherscanGoerliBaseUrl = 'https://api-goerli.etherscan.io/api';

    /**
     * Get default ecoyn payment provider associated with this blockchain.
     *
     * @return AbstractPaymentProvider
     */
    public static function getPaymentProvider(): AbstractPaymentProvider
    {
        return new Provider();
    }

    /**
     * Get unique blockchain identifier.
     *
     * @return string
     */
    public static function getBlockchainName(): string
    {
        return 'ETH';
    }

    /**
     * Get JavaScript control for payment display.
     *
     * This control MUST extend package/quiqqer/payment-crypto/bin/js/frontend/controls/PaymentDisplay
     *
     * @return string
     */
    public static function getPaymentDisplayJavaScriptControl(): string
    {
        return 'package/quiqqer/payment-crypto-eth/bin/js/frontend/controls/PaymentDisplay';
    }

    /**
     * Get transaction data from blockchain.
     *
     * @param string $txId
     * @param CryptoPaymentType $CryptoPaymentType
     * @return BlockchainTransaction
     *
     * @throws QUI\ERP\Payments\Crypto\Exception\Exception
     */
    public static function getTransaction(string $txId, CryptoPaymentType $CryptoPaymentType): BlockchainTransaction
    {
        $apiConfig = self::getApiConfig();
        $etherscanApiKey = $apiConfig['etherscan_api_key'];

        try {
            $txData = CryptoPaymentsUtils::request(
                self::getEtherscanBaseUrl(),
                CryptoPaymentsUtils::API_REQUEST_METHOD_GET,
                [
                    'apikey' => $etherscanApiKey,
                    'module' => 'proxy',
                    'action' => 'eth_getTransactionByHash',
                    'txhash' => $txId
                ],
                [
                    'Content-Type: application/json'
                ]
            );
        } catch (\Exception $Exception) {
            QUI\System\Log::writeException($Exception);

            throw new TransactionNotFoundException(
                'Etherscan - Cannot check status of tx "' . $txId . '". Transaction GET request failed.'
            );
        }

        // check tx
        $txData = json_decode($txData, true);

        if (empty($txData['result'])) {
            throw new TransactionNotFoundException(
                'Etherscan - Cannot check status of tx "' . $txId . '". Transaction not found no blockchain.'
            );
        }

        if (
            !isset($txData['result']['from'])
            || !isset($txData['result']['type'])
            || !isset($txData['result']['value'])
            || !isset($txData['result']['to'])
        ) {
            throw new TransactionNotFoundException(
                'Etherscan - Cannot check status of tx "' . $txId . '". Transaction data is missing fields.'
            );
        }

        $status = false;

        try {
            $status = CryptoPaymentsUtils::request(
                self::getEtherscanBaseUrl(),
                CryptoPaymentsUtils::API_REQUEST_METHOD_GET,
                [
                    'apikey' => $etherscanApiKey,
                    'module' => 'transaction',
                    'action' => 'gettxreceiptstatus',
                    'txhash' => $txId
                ],
                [
                    'Content-Type: application/json'
                ]
            );

            $status = json_decode($status, true);
        } catch (\Exception $Exception) {
            QUI\System\Log::writeException($Exception);
        }


        $from = $txData['result']['from'];
        $to = $txData['result']['to'];
        $input = $txData['result']['input'];

        switch ($CryptoPaymentType->getCryptoToken()->getSymbol()) {
            // Native currncy
            case 'ETH':
                $amount = (int)hexdec($txData['result']['value']); // wei
                break;

            // Non-native currency (token)
            default:
                if (mb_strlen($input) !== 138) {
                    throw new QUI\ERP\Payments\Crypto\Exception\Exception(
                        'Etherscan - Cannot check status of tx "' . $txId . '". Input is invalid.',
                        0,
                        [
                            'input' => $input
                        ]
                    );
                }

                $to = '0x' . mb_substr($input, 34, 40);
                $amount = (int)hexdec(mb_substr($input, 74));   // wei
        }

        $amount = $amount / 1000000000000000000;     // eth

        $BlockchainTransaction = new BlockchainTransaction(
            $txId,
            $from,
            $to,
            $amount,
            $CryptoPaymentType->getCurrency()
        );

        if (
            $status &&
            (isset($status['result']['status']) && (int)$status['result']['status'] === 1) ||
            (isset($status['message']) && $status['message'] === 'OK')
        ) {
            $BlockchainTransaction->setStatus(TransactionsHandler::STATUS_COMPLETE);
        }

        return $BlockchainTransaction;
    }

    /**
     * @return array
     */
    protected static function getApiConfig(): array
    {
        try {
            $Conf = QUI::getPackage('quiqqer/payment-crypto-eth')->getConfig();
        } catch (\Exception $Exception) {
            QUI\System\Log::writeException($Exception);
            return [];
        }

        return $Conf->getSection('api');
    }

    /**
     * Check if this blockchain provider is set up for payment.
     *
     * @return bool
     */
    public static function isSetUp(): bool
    {
        $apiConfig = self::getApiConfig();

        if (empty($apiConfig['etherscan_api_key'])) {
            return false;
        }

        return true;
    }

    /**
     * Shall the Goerli test network be used?
     *
     * @return bool
     */
    public static function isTestMode(): bool
    {
        $apiConfig = self::getApiConfig();
        return !empty($apiConfig['test_mode']);
    }

    /**
     * Get number of seconds for the SYNCHRONOUS validation of a transaction on this blockchain (in the order process).
     *
     * @return int - Seconds
     */
    public static function getSynchronousTransactionValidationTime(): int
    {
        $apiConfig = self::getApiConfig();

        if (empty($apiConfig['validation_time_sync'])) {
            return 30; // fallback
        }

        return (int)$apiConfig['validation_time_sync'];
    }

    /**
     * Get maximum number of seconds a blockchain transaction may be tried to be validated.
     *
     * @return int - Hours
     */
    public static function getMaxTransactionValidationTime(): int
    {
        $apiConfig = self::getApiConfig();

        if (empty($apiConfig['max_validation_time'])) {
            return 72; // fallback
        }

        return (int)$apiConfig['max_validation_time'];
    }

    /**
     * @return string
     */
    protected static function getEtherscanBaseUrl(): string
    {
        return self::isTestMode() ? self::$etherscanGoerliBaseUrl : self::$etherscanBaseUrl;
    }
}
